from collections import OrderedDict
from .constants import EOT_IDENTIFIER


class Encoder(object):

    """
        Encodes payload to the bencode format.
    """

    def __init__(self, payload):
        self._payload = payload

    def encode(self):
        return self._encode_payload(self._payload)

    def _encode_payload(self, payload):
        if isinstance(payload, str):
            return self._from_string(payload)
        elif isinstance(payload, int):
            return self._from_int(payload)
        elif isinstance(payload, list):
            return self._from_list(payload)
        elif isinstance(payload, bytes):
            return self._from_bytes(payload)
        elif isinstance(payload, dict) or isinstance(payload, OrderedDict):
            return self._from_dict(payload)
        else:
            data_type = type(payload).__name__
            err = f'[error] Cannot encode object of type {data_type}'
            raise TypeError(err)

    def _from_dict(self, _dict):
        # we use a bytearray because it's impossible to
        # concatenate bytearray types to strings

        encoded_dict = bytearray('d', 'utf-8')
        for key, value in _dict.items():
            k = self._encode_payload(key)
            v = self._encode_payload(value)

            encoded_dict += k
            encoded_dict += v

        encoded_dict += EOT_IDENTIFIER
        return encoded_dict

    def _from_list(self, _list):
        # we use a bytearray because it's impossible to
        # concatenate bytearray types to strings

        encoded_list = bytearray('l', 'utf-8')
        encoded_list += b''.join([
            self._encode_payload(_list[idx])
            for idx in range(len(_list))
        ])
        encoded_list += EOT_IDENTIFIER
        return encoded_list

    def _from_int(self, _int):
        return bytes(f'i{str(_int)}e', 'utf-8')

    def _from_string(self, string):
        return bytes(f'{len(string)}:{string}', 'utf-8')

    def _from_bytes(self, _bytes):
        # copied from pieces.bencoding.Encoder
        result = bytearray()
        result += str.encode(str(len(_bytes)))
        result += b':'
        result += _bytes
        return result



from collections import OrderedDict
from .constants import (
    DICT_IDENTIFIER, INT_IDENTIFIER,
    LIST_IDENTIFIER, STRING_DELIMITER,
    EOT_IDENTIFIER
)


class Decoder(object):

    """
        Decodes a bencoded object into
        its corresponding python representation.
    """

    def __init__(self, payload):

        if not isinstance(payload, bytes):
            _type = type(payload).__name__
            err = f'[error] Cannot decode payload. Expected bytes, got {_type}'
            raise TypeError(err)

        self._payload = payload
        self._index = 0

    def decode(self):
        char = self._opening_char

        if char == DICT_IDENTIFIER:
            self._index += 1
            return self._to_dict()
        elif char == INT_IDENTIFIER:
            self._index += 1
            return self._to_int()
        elif char == LIST_IDENTIFIER:
            self._index += 1
            return self._to_list()
        elif char is None:
            return
        elif char in b'0123456789':
            return self._to_string()

    def _to_dict(self):

        # Note: an OrderedDict is being used in this case because we need
        # the 'info' dictionaries to be identical in order to produce the
        # correct sha1 hash that will be used to make a request to the tracker.

        _dict = OrderedDict()

        # loop through the payload until we reach 'e'
        while self._payload[self._index: self._index + 1] != EOT_IDENTIFIER:
            # decode whatever is at that index
            key = self.decode()
            value = self.decode()

            # populate the final list
            _dict[key] = value

        # when we're done, increment the index.
        # if it's bigger or equal to the length of the payload
        # the next call to self._opening_char will return None
        # and the program will exit
        self._index += 1
        return _dict

    def _to_list(self):
        _list = []
        # loop through the payload until we reach 'e'
        while self._payload[self._index:self._index+1] != EOT_IDENTIFIER:
            # decode whatever is at that index
            element = self.decode()

            # populate the final list
            _list.append(element)

        # when we're done, increment the index.
        # if it's bigger or equal to the length of the payload
        # the next call to self._opening_char will return None
        # and the program will exit

        self._index += 1
        return _list

    def _to_int(self):
        # access EOT_IDENTIFIER (ie 'e')
        end_token_index = self._payload.index(EOT_IDENTIFIER, self._index)

        # read whatever is between 'i' and 'e' and cast that to an int
        _int = int(self._payload[self._index:end_token_index])

        # update index to start after EOT_IDENTIFIER
        self._index = end_token_index + 1

        return _int

    def _to_string(self):
        # get index of delimiter
        str_delim_index = self._payload.index(STRING_DELIMITER, self._index)

        # number of characters to read
        chars_to_read = int(self._payload[self._index:str_delim_index])

        # set the starting index at delimiter + 1 (ie the char after :)
        start_index = str_delim_index + 1

        # read `char_to_read` characters starting from delimiter index
        data = self._payload[start_index:start_index + chars_to_read]

        # update index to start at end of string
        self._index = start_index + chars_to_read

        return data

    @property
    def _opening_char(self):
        if self._index >= len(self._payload):
            # we've reached the end of the payload
            return None
        return self._payload[self._index:self._index + 1]
